from typing import Any
from hashlib import sha256


def generate_sha256(obj: Any) -> str:
    """Generates a SHA-256 hash of an object.
    
    Args:
        obj (Any): Input object.
    
    Returns:
        (str): SHA-256 representation of `obj`.
    """
    hasher = sha256()
    # NOTE: Prevents different hashes caused by different memory addresses
    hasher.update(repr(obj))
    hash = hasher.hexdigest()
    return hash


def verify_sha256(obj: Any, hash: str) -> bool:
    """Checks if the SHA-256 hash of an object matches a given hash.
    
    Args:
        obj (Any): Input object to check.
        hash (str): Ground truth hash to check.
    
    Returns:
        (bool): `True` if `obj` hash and `hash` are the same, `False`
            otherwise.
    """
    obj_hash = generate_sha256(obj)
    return obj_hash == hash


def generate_sha256_from_file(file: str, block_size: int = 65_536) -> str:
    """Generates a SHA-256 hash of a file.
    
    Args:
        file (str): Input file.
        block_size (int): Block size used to read the file.
    
    Returns:
        (str): SHA-256 hash of the file's content.
    """
    # NOTE: block_size should ideally be multiple of the byte digest block
    hasher = sha256()

    with open(file, "rb") as f:
        for block in iter(lambda: f.read(block_size), b""):
            hasher.update(block)

    hash = hasher.hexdigest()
    return hash


def verify_sha256_from_file(
        file: str,
        hash: str,
        block_size: int = 65_536
) -> bool:
    """Checks if the SHA-256 hash of a file matches a given hash.
    
    Args:
        file (str): Input file to check.
        hash (str): Ground truth hash to check.
        block_size (int): Block size used to read the file.
    
    Returns:
        (bool): `True` if the file's hash and `hash` are the same, `False`
            otherwise.
    """
    file_hash = generate_sha256_from_file(file, block_size=block_size)
    return file_hash == hash



import sys
import argparse
from .cmd import sndls
from ..utils.fmt import (
    printc_exit as print_exit,
    exit_warning
)
from sndls import (
    __description__,
    __version_repr__
)


def get_parser() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description=__description__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        allow_abbrev=False
    )
    parser.add_argument(
        "input",
        type=str,
        nargs="?",
        default=".",
        help="input audio file, .csv file or folder containing audio files"
    )
    parser.add_argument(
    "-e", "--extension",
        type=str,
        nargs="+",
        default=[".wav"],
        help="audio file extension(s)"
    )
    parser.add_argument(
        "-r", "--recursive",
        action="store_true",
        help="search input folder recursively"
    )
    parser.add_argument(
        "-d", "--dtype",
        choices=["float32", "float64"],
        default="float32",
        help="data type used to read audio files and calculate stats"
    )
    parser.add_argument(
        "-m", "--meta",
        action="store_true",
        help="show metadata information only"
    )
    parser_filter_select = parser.add_mutually_exclusive_group()
    parser_filter_select.add_argument(
        "-f", "--filter",
        type=str,
        help="filter files meeting a certain condition"
    )
    parser_filter_select.add_argument(
        "-s", "--select",
        type=str,
        help="select files meeting a certain condition"
    )
    parser_hash = parser.add_mutually_exclusive_group()
    parser_hash.add_argument(
        "--sha256",
        action="store_true",
        help="compute sha256 hash"
    )
    parser_hash.add_argument(
        "--sha256-short",
        action="store_true",
        help="compute sha256 hash and print only last 8 characters"
    )
    parser.add_argument(
        "--csv",
        type=str,
        help="save output to a .csv file"
    )
    parser.add_argument(
        "--sample",
        type=float,
        help=(
            "number of files to randomly sample from all input files. If the "
             "value is between 0.0 and 1.0, it will be interpreted as a "
             "percentage"
        )
    )
    parser.add_argument(
        "--summary",
        action="store_true",
        help="print summary only"
    )
    parser.add_argument(
        "--silent-thresh",
        type=float,
        default=-80.0,
        help="root mean square (RMS) threshold in decibels below which a file "
             "is considered silent"
    )
    parser.add_argument(
        "--fft-size",
        type=int,
        choices=[2048, 1024, 512],
        default=2048,
        help="fft size for frequency spectrum"
    )
    parser.add_argument(
        "--hop-size",
        type=int,
        choices=[512, 256, 128],
        default=512,
        help="hop size for frame shifts"
    )
    parser.add_argument(
        "--spectral-rolloff",
        type=float,
        help=(
            "computes the mean spectral rolloff in hertz for a given "
            "cumulative energy percentage between 0.0 and 1.0"
        )
    )
    parser.add_argument(
        "--spectral-rolloff-detail",
        action="store_true",
        help="shows spectral rollof in min ≤ mean ≤ max format"
    )
    parser.add_argument(
        "-p", "--post-action",
        choices=["cp", "mv", "rm", "mv+sp", "cp+sp"],
        help=(
            "action to execute after listing all files (cp=copy, mv=move, "
            "rm=remove, sp=split)"
        )
    )
    parser.add_argument(
        "--post-action-output",
        type=str,
        help=(
            "post action output (required for --post-action "
            "{cp,mv,cp+sp,mv+sp})"
        )
    )
    parser.add_argument(
        "--post-action-preserve-subfolders",
        action="store_true",
        help=(
            "preserve subfolder structure with respect to --input when copying"
            " or moving files using --post-action"
        )
    )
    parser.add_argument(
        "--post-action-num-splits",
        type=int,
        help="number of partitions (required for --post-action {cp+sp,mv+sp})"
    )
    parser.add_argument(
        "--post-action-split-dirname",
        type=str,
        default="split_",
        help="split folder name (only valid if --post-action {cp+sp,mv+sp})"
    )
    parser.add_argument(
        "--max-fname-chars",
        type=int,
        default=40,
        help="maximum filename characters to display"
    )
    parser.add_argument(
        "--csv-input-file-col",
        type=str,
        default="file",
        help="name of the column containing files if the input is a .csv file"
    )
    parser.add_argument(
        "--max-duration",
        type=float,
        default=60 * 60 * 3,
        help="skip reading audio files larger than this duration in seconds"
    )
    parser.add_argument(
        "--skip-invalid-files",
        action="store_true",
        help="skip and report files that cannot be parsed"
    )
    parser.add_argument(
        "--preload",
        type=str,
        help="preload .csv, .tsv or .txt file to use as input for "
             "--filter/--select"
    )
    parser.add_argument(
        "--preload-truncate-ragged-lines",
        action="store_true",
        help="truncate ragged lines when found in preloaded files"
    )
    parser.add_argument(
        "--preload-has-header",
        action="store_true",
        help="if enabled, the first row of --preload will be ignored"
    )
    parser.add_argument(
        "--csv-overwrite",
        action="store_true",
        help="overwrites .csv file if --csv is enabled and file already exists"
    )
    parser.add_argument(
        "-u", "--unattended",
        action="store_true",
        help="perform --post-action without user confirmation"
    )
    parser.add_argument(
        "--random-seed",
        type=int,
        default=1234,
        help="random seed used if --sample is enabled"
    )

    return parser


def main() -> None:
    # Print tool version
    if (len(sys.argv) == 2 and sys.argv[1] in ("-v", "--version")):
        print_exit(__version_repr__, code=0)
    
    # NOTE: If no input is given, then current directory is assumed
    parser = get_parser()
    args = parser.parse_args()

    try:
        sndls(args)
    
    except KeyboardInterrupt:
        exit_warning("Process terminated by the user")



import os
import csv
import random
import shutil
import numpy as np
import polars as pl
from copy import deepcopy
from time import perf_counter
from decimal import Decimal
from numbers import Number
from argparse import Namespace
from tqdm import tqdm
from typing import List
from ..utils.config import (
    get_allowed_audio_file_extensions,
    get_sppbar_color
)
from ..utils.io import (
    ask_confirmation,
    get_dir_files,
    read_audio,
    read_audio_metadata
)
from ..utils.collections import flatten_nested_list
from ..utils.fmt import (
    bytes_to_str,
    exit_error,
    exit_warning,
    printc as print,
    print_error,
    print_warning,
    time_to_str
)
from ..utils.guards import is_file_with_ext
from ..utils.hash import generate_sha256_from_file
from ..utils.audio import (
    is_anomalous,
    is_clipped,
    is_silent,
    peak_db,
    rms_db,
    spectral_rolloff
)


def _matches_filter(
        data: dict,
        preload: pl.DataFrame,
        expr: str
) -> bool:
    """Matches a filter expression against a set of file specifications.
    
    Args:
        data (dict): Audio file specifications.
        preload (pl.DataFrame): Preloaded data.
        expr (str): Filter expression.
    
    Returns:
        bool: `True` of the filter matches the contents of `data`, `False`
            otherwise.
    """
    try:
        # Set constrained globals and locals
        if preload is not None:
            data["preload"] = preload

        result = eval(expr, {}, deepcopy(data))

        if not isinstance(result, bool):
            raise ValueError("Invalid return type")
    
    except Exception as e:
        fields_repr = ", ".join(k for k in data)

        exit_error(
            f"Invalid --filter/--select expression '{expr}': {e}.\n"
            "Only python expressions returning a bool value are valid. To "
            "create a filter expression you can access any of the following "
            f"fields of each file: {fields_repr}. "
            "Please look at the repository README.md for further details",
            writer=tqdm
        )
    
    return result


def _preload_file(
        file: str,
        has_header: bool = False,
        truncate_ragged_lines: bool = False
) -> pl.DataFrame:
    """Preloads a file in memory to be used with --filter/--select option.
    
    Args:
        file (str): File to be loaded in memory as a `pl.DataFrame`.
        has_header (bool): If `True`, the first column of the preloaded file
            is assumed to be a header and will be ignored.
    """
    if not os.path.isfile(file):
        exit_error(f"--preload file '{file}' not found")
    
    elif not is_file_with_ext(file, ext=[".csv", ".tsv", ".txt"]):
        exit_error("--preload option only supports .csv, .tsv or .txt files")
    
    elif is_file_with_ext(file, ext=".csv"):
        separator = ","
    
    elif is_file_with_ext(file, ext=".txt"):
        separator = " "
    
    elif is_file_with_ext(file, ext=".tsv"):
        separator = "\t"
    
    else:
        raise AssertionError
    
    try:
        preload = pl.read_csv(
            file,
            separator=separator,
            has_header=has_header,
            truncate_ragged_lines=truncate_ragged_lines
        )
        
    except pl.exceptions.ComputeError as e:
        exit_error(
            f"The following error occurred while opening '{file}': {e}\n\n"
            "If the issue was caused by 'truncate_ragged_lines', please "
            "consider using --preload-truncate-ragged-lines"
        )
    
    return preload


def _audio_file_meta_repr_from_dict(data: dict, max_fname_chars: int) -> str:
    """Creates a printable string representation of a set of audio file
    specifications.

    !!! note
        Differently from `_audio_file_repr_from_dict`, the specifications
        in `data` of this method contain exclusively metadata information.
    
    Args:
        data (dict): Audio data.
        max_fname_chars (int): Maximum name of characters from the file path
            to be printed to the terminal.
        
    Returns:
        str: `str` representation of the audio file specifications.
    """
    # Get filename repr
    filename_repr = (
        f"...{data['file'][-max_fname_chars:]}"
        if len(data["file"]) > max_fname_chars + 3 else data["file"]
    ).ljust(max_fname_chars + 3)

    # Get length repr
    num_samples_repr = str(data["num_samples_per_channel"])

    if len(num_samples_repr) > 10:
        f"{Decimal(num_samples_repr):.5e}"
    
    if data["is_invalid"]:
        len_repr = "-".rjust(20)

    else:
        len_repr = (
            num_samples_repr
            + "x" + str(data["num_channels"])
            + "@" + str(data["fs"]) + "hz"
        ).rjust(20)

    # Get mem repr
    mem_repr = bytes_to_str(data["size_bytes"]).rjust(7)

    # Get format repr
    data_fmt = "-" if data["fmt"] is None else data["fmt"]
    data_subtype = "-" if data["subtype"] is None else data["subtype"]

    fmt_repr = data_fmt.ljust(4) + " " + data_subtype.ljust(14)

    # Assemble representation
    repr = f"{filename_repr} {mem_repr} {fmt_repr} {len_repr}"

    if data["is_invalid"]:
        repr = f"<error>{repr}</error>"

    return repr


def _audio_file_repr_from_dict(
        data: dict,
        max_fname_chars: int,
        abbrev_hash: bool
) -> str:
    """Creates a printable string representation of a set of audio file
    specifications.
    
    !!! note
        Differently from `_audio_file_meta_repr_from_dict`, the specifications
        in `data` contain metadata information and audio based statistics.
    
    Args:
        data (dict): Audio data.
        max_fname_chars (int): Maximum name of characters from the file path
            to be printed to the terminal.
        
    Returns:
        str: `str` representation of the audio file specifications.
    """
    # Get filename repr
    filename_repr = (
        f"...{data['file'][-max_fname_chars:]}"
        if len(data["file"
                    ]) > max_fname_chars + 3 else data["file"]
    ).ljust(max_fname_chars + 3)

    # Get length repr
    num_samples_repr = str(data["num_samples_per_channel"])

    if len(num_samples_repr) > 10:
        f"{Decimal(num_samples_repr):.5e}"
    
    if data["is_invalid"]:
        len_repr = "-".rjust(20)
    
    else:
        len_repr = (
            num_samples_repr
            + "x" + str(data["num_channels"])
            + "@" + str(data["fs"]) + "hz"
        ).rjust(20)

    # Get mem repr
    mem_repr = bytes_to_str(data["size_bytes"]).rjust(7)

    # Get format repr
    data_fmt = "-" if data["fmt"] is None else data["fmt"]
    data_subtype = "-" if data["subtype"] is None else data["subtype"]

    fmt_repr = data_fmt.ljust(4) + " " + data_subtype.ljust(14)

    # Audio stats repr
    if data["is_invalid"]:
        db_repr = "- dBrms".rjust(16) + " " + "- dBpeak".rjust(16)
    
    else:
        if all(
            s in data
            for s in (
                "spectral_rolloff_min",
                "spectral_rolloff",
                "spectral_rolloff_max")
        ):
            db_repr = " ".join(
                [
                    f"{r:.1f}dBrms:{idx}".rjust(16)
                    + f"{p:.1f}dBpeak:{idx}".rjust(16)
                    + "  "
                    + f"{s_min / 1000.0:.1f}".rjust(4) 
                    + " ≤"
                    + f"{s_avg / 1000.0:.1f}".rjust(4)
                    + " ≤"
                    + f"{s_max / 1000.0:.1f}".rjust(4)
                    + f"kHz:{idx}".ljust(6)
                    for idx, (r, p, s_min, s_avg, s_max)
                    in enumerate(
                        zip(
                            data["rms_db"],
                            data["peak_db"],
                            data["spectral_rolloff_min"],
                            data["spectral_rolloff"],
                            data["spectral_rolloff_max"]
                        )
                    )
                ]
            )

        elif "spectral_rolloff" in data:
            db_repr = " ".join(
                [
                    f"{r:.1f}dBrms:{idx}".rjust(16)
                    + f"{p:.1f}dBpeak:{idx}".rjust(16)
                    + f"{s / 1000.0:.1f}kHz:{idx}".rjust(12)
                    for idx, (r, p, s)
                    in enumerate(
                        zip(
                            data["rms_db"],
                            data["peak_db"],
                            data["spectral_rolloff"]
                        )
                    )
                ]
            )
        
        else:
            db_repr = " ".join(
                [
                    f"{r:.1f}dBrms:{idx}".rjust(16)
                    + f"{p:.1f}dBpeak:{idx}".rjust(16)
                    for idx, (r, p)
                    in enumerate(zip(data["rms_db"], data["peak_db"]))
                ]
            )

    # Assemble representation
    repr = f"{filename_repr} {mem_repr} {fmt_repr} {len_repr} {db_repr}"

    if "sha256" in data:
        sha256 = data["sha256"][-8:] if abbrev_hash else data["sha256"]
        repr = (
            f"{filename_repr}  {sha256} {mem_repr} {fmt_repr} {len_repr} "
            f"{db_repr}"
        )

    if (
        data["is_clipped"]
        or data["is_anomalous"]
        or data["is_silent"]
        or data["is_invalid"]
    ):
        repr = f"<error>{repr}</error>"

    return repr


def _create_dir_if_missing(dir: str) -> None:
    """Creates a folder if it does not exist.
    
    Args:
        dir (str): Folder to be created.
    """
    if not os.path.isdir(dir):
        print(f"Creating post action output folder '{dir}'")

        try:
            os.makedirs(dir)
        
        except Exception as e:
            exit_error(f"An unexpected error ocurred: {e}")


def _perform_post_action(files: List[str], args: Namespace) -> None:
    """Perform post actions such as copying, moving or deleting files that
    match a certain filter.
    
    Args:
        files (List[str]): List of targeted files.
        args (Namespace): Main namespace containing user provided input.
    """
    if args.post_action == "cp":
        output = args.post_action_output
        print(f"\n{len(files)} file(s) will be copied to '{output}'")

        if not args.unattended:
            ask_confirmation()
        
        _create_dir_if_missing(args.post_action_output)

        copied_files = 0

        # Copy files and warn user if a file already exists
        for f in tqdm(
            files,
            desc="Copying files",
            leave=False,
            unit="file",
            colour=get_sppbar_color()
        ):
            try:
                if args.recursive and args.post_action_preserve_subfolders:
                    # Get relative path
                    dst = f.replace(args.input, "", 1)
                    dst = dst.split(os.sep)  # Avoid double-slashes xplatform
                    dst = os.path.join(args.post_action_output, *dst)

                    # Create subfolders if structure does not exist yet
                    os.makedirs(os.path.dirname(dst), exist_ok=True)
                
                else:
                    dst = os.path.join(output, os.path.basename(f))

                if os.path.exists(dst):
                    print_warning(
                        f"File '{dst}' already exists",
                        writer=tqdm
                    )
                    continue

                else:
                    shutil.copy(f, dst)
                    copied_files += 1
            
            except Exception as e:
                print_warning(
                    f"An error ocurred while copying '{f}' to '{dst}': {e}",
                    writer=tqdm
                )
            
        print_fn = (print_warning if copied_files != len(files) else print)
        print_fn(f"{copied_files}/{len(files)} file(s) copied to '{output}'")
    
    elif args.post_action == "mv":
        output = args.post_action_output
        print(f"\n{len(files)} file(s) will be moved to '{output}'")

        if not args.unattended:
            ask_confirmation()
        
        _create_dir_if_missing(args.post_action_output)
    
        # Move files and warn user if a file already exists
        moved_files = 0

        # Move audio files and warn user if a file already exists
        for f in tqdm(
            files,
            desc="Moving audio files",
            leave=False,
            unit="file",
            colour=get_sppbar_color()
        ):
            try:
                if args.recursive and args.post_action_preserve_subfolders:
                    # Get relative path
                    dst = f.replace(args.input, "", 1)
                    dst = dst.split(os.sep)  # Avoid double-slashes xplatform
                    dst = os.path.join(args.post_action_output, *dst)

                    # Create subfolders if structure does not exist yet
                    os.makedirs(os.path.dirname(dst), exist_ok=True)
                
                else:
                    dst = os.path.join(output, os.path.basename(f))

                if os.path.exists(dst):
                    print_warning(f"File '{dst}' already exists", writer=tqdm)
                    continue
                    
                else:
                    shutil.move(f, dst)
                    moved_files += 1
            
            except Exception as e:
                print_warning(
                    f"An error ocurred while moving '{f}' to '{dst}': {e}",
                    writer=tqdm
                )
            
        print_fn = (print_warning if moved_files != len(files) else print)
        print_fn(f"{moved_files}/{len(files)} file(s) moved to '{output}'")
    
    elif args.post_action == "rm":
        print(
            f"\n{len(files)} file(s) will be deleted. This action cannot be "
            "undone"
        )

        if not args.unattended:
            ask_confirmation()
        
        # Delete files
        for f in tqdm(
            files,
            desc="Deleting files",
            leave=False,
            unit="file",
            colour=get_sppbar_color()
        ):
            try:
                os.remove(f)
            
            except Exception as e:
                print_warning(
                    f"An error ocurred while deleting '{f}': {e}",
                    writer=tqdm
                )
    
    elif args.post_action == "cp+sp":
        # Check splits are possible without 0 files in any of them
        if len(files) < args.post_action_num_splits:
            exit_error(
                "The number of requested splits "
                f"({args.post_action_num_splits}) is bigger than the number of"
                f" file(s) ({len(files)})"
            )

        print(
            f"\n{len(files)} file(s) will be copied and split into "
            f"{args.post_action_num_splits} partition(s)"
        )

        if not args.unattended:
            ask_confirmation()
        
        _create_dir_if_missing(args.post_action_output)

        # Shuffle files and create splits specs
        random.seed(args.random_seed)
        random.shuffle(files)
        splits = np.array_split(files, args.post_action_num_splits)
        split_dirname_zfill = len(str(args.post_action_num_splits - 1))
        copied_files = 0

        for split_idx, split_data in enumerate(splits):
            split_files = split_data.tolist()
            split_dir = os.path.join(
                args.post_action_output,
                f"{args.post_action_split_dirname}"
                f"{str(split_idx).zfill(split_dirname_zfill)}"
            )
            _create_dir_if_missing(split_dir)

            # Copy files and warn user if a file already exists
            for f in tqdm(
                split_files,
                desc=(
                    f"Copying files of partition {split_idx + 1}/"
                    f"{args.post_action_num_splits}"
                ),
                leave=False,
                unit="file",
                colour=get_sppbar_color()
            ):
                try:
                    if args.recursive and args.post_action_preserve_subfolders:
                        # Get relative path
                        dst = f.replace(args.input, "", 1)
                        dst = dst.split(os.sep)  # Avoid double-slashes
                        dst = os.path.join(split_dir, *dst)

                        # Create subfolders if structure does not exist yet
                        os.makedirs(os.path.dirname(dst), exist_ok=True)
                    
                    else:
                        dst = os.path.join(split_dir, os.path.basename(f))
                    
                    if os.path.exists(dst):
                        print_warning(
                            f"File '{dst}' already exists",
                            writer=tqdm
                        )
                        continue

                    else:
                        shutil.copy(f, dst)
                        copied_files += 1
                
                except Exception as e:
                    print_warning(
                        f"An error ocurred while copying '{f}' to '{dst}': {e}",
                        writer=tqdm
                    )

        print_fn = (print_warning if copied_files != len(files) else print)
        print_fn(
           f"{copied_files}/{len(files)} file(s) copied to "
           f"'{args.post_action_output}'"
        )
    
    elif args.post_action == "mv+sp":
        # Check splits are possible without 0 files in any of them
        if len(files) < args.post_action_num_splits:
            exit_error(
                "The number of requested splits "
                f"({args.post_action_num_splits}) is bigger than the number of"
                f" file(s) ({len(files)})"
            )

        print(
            f"\n{len(files)} file(s) will be moved and split into "
            f"{args.post_action_num_splits} partition(s)"
        )

        if not args.unattended:
            ask_confirmation()
        
        _create_dir_if_missing(args.post_action_output)

        # Create splits specs
        random.seed(args.random_seed)
        random.shuffle(files)
        splits = np.array_split(files, args.post_action_num_splits)
        split_dirname_zfill = len(str(args.post_action_num_splits - 1))
        moved_files = 0

        for split_idx, split_data in enumerate(splits):
            split_files = split_data.tolist()
            split_dir = os.path.join(
                args.post_action_output,
                f"{args.post_action_split_dirname}"
                f"{str(split_idx).zfill(split_dirname_zfill)}"
            )
            _create_dir_if_missing(split_dir)

            # Copy files and warn user if a file already exists
            for f in tqdm(
                split_files,
                desc=(
                    f"Moving files of partition {split_idx + 1}/"
                    f"{args.post_action_num_splits}"
                ),
                leave=False,
                unit="file",
                colour=get_sppbar_color()
            ):
                try:
                    if args.recursive and args.post_action_preserve_subfolders:
                        # Get relative path
                        dst = f.replace(args.input, "", 1)
                        dst = dst.split(os.sep)  # Avoid double-slashes
                        dst = os.path.join(split_dir, *dst)

                        # Create subfolders if structure does not exist yet
                        os.makedirs(os.path.dirname(dst), exist_ok=True)
                    
                    else:
                        dst = os.path.join(split_dir, os.path.basename(f))
                    
                    if os.path.exists(dst):
                        print_warning(
                            f"File '{dst}' already exists",
                            writer=tqdm
                        )
                        continue

                    else:
                        shutil.move(f, dst)
                        moved_files += 1
                
                except Exception as e:
                    print_warning(
                        f"An error ocurred while moving '{f}' to '{dst}': {e}",
                        writer=tqdm
                    )

        print_fn = (print_warning if moved_files != len(files) else print)
        print_fn(
           f"{moved_files}/{len(files)} file(s) moved to "
           f"'{args.post_action_output}'"
        )

    else:
        raise AssertionError


def sndls(args: Namespace) -> None:
    """Main routine triggered by the `sndls` command.
    
    Args:
        args (Namespace): Main namespace containing user provided input.
    """
    # Check file extensions
    for ext in args.extension:
        if ext not in get_allowed_audio_file_extensions():
            exit_error(
                "Currently only the following extensions are supported: "
                f"{', '.join(get_allowed_audio_file_extensions())}"
            )
    
    # Check incompatible args that are not handled by mutually exclusive groups
    if args.meta and (
        args.sha256
        or args.sha256_short
        or args.csv
        or args.filter
        or args.select 
        or args.spectral_rolloff
    ):
        exit_error(
            "--meta not allowed with: --sha256, --sha256-short, --csv, "
            "--filter, --select, --spectral-rolloff"
        )
    
    # Check spectral-rolloff if enabled
    if (
        args.spectral_rolloff is not None
        and (args.spectral_rolloff > 1.0 or args.spectral_rolloff < 0.0)
    ):
        exit_error("--spectral-rolloff should be a value between 0.0 and 1.0")
    
    # Check csv does not exist already if it should be written
    if args.csv and os.path.isfile(args.csv) and not args.csv_overwrite:
        exit_error(
            f"'{args.csv}' already exists. Please choose a different filename "
            "or use --csv-overwrite to allow overwriting existing files"
        )
    
    if args.sample is not None and args.sample <= 0.0:
        exit_error(
            "--sample should be 1 or greater to sample a concrete number of "
            "samples or between 0.0 and 1.0 to sample a percentage of all "
            "samples"
        )
    
    # Preload file if requested
    if args.preload is not None:
        preload = _preload_file(
            file=args.preload,
            has_header=args.preload_has_header,
            truncate_ragged_lines=args.preload_truncate_ragged_lines
        )
    
    else:
        preload = None
    
    # Get file(s)
    if is_file_with_ext(file=args.input, ext=args.extension):
        files = [args.input]
    
    elif is_file_with_ext(file=args.input, ext=".csv"):
        # Read file and check if the col column exists
        df = pl.read_csv(args.input)

        if args.csv_input_file_col not in df.columns:
            exit_error(
                f"No '{args.csv_input_file_col}' column found in the provided "
                f".csv file '{args.input}'"
            )

        # Read files and check all exist since col can contain garbage values
        files = df[args.csv_input_file_col].to_list()

        for row_idx, file in enumerate(tqdm(
            files,
            desc=f"Verifying '{args.csv_input_file_col}' column data",
            colour=get_sppbar_color(),
            leave=False,
            unit="row"
        )):
            if not is_file_with_ext(file=file, ext=args.extension):
                # NOTE: +2 because the count starts from 1 and the header
                # row is skipped in the count
                exit_error(
                    f"Row #{row_idx + 2} of '{args.csv_input_file_col}' column"
                    f" contains an invalid filename '{file}'. Please check "
                    "that such a file exists, has a valid --extension option, "
                    "and can be reached"
                )
        
    elif os.path.isdir(args.input):
        # Show progress bar in case folder is too big
        with tqdm(
            total=1,
            desc="Fetching files... This may take some time for large folders",
            bar_format="{desc}",
            leave=False
        ) as pbar:
            pbar.update(1)
            files = get_dir_files(
                dir=args.input,
                ext=args.extension,
                recursive=args.recursive
            )
        
    else:
        exit_error(f"Invalid input file or folder '{args.input}'")

    # Check folder is not empty
    if len(files) == 0:
        if not args.recursive:
            exit_warning(
                f"0 audio files found in '{args.input}'. Use --recursive or -r"
                " if you intended to perform a recursive search"
            )
        else:
            exit_warning(f"0 audio files found in '{args.input}'")

    # Check splits are provided
    if (
        args.post_action in ("mv+sp", "cp+sp")
        and args.post_action_num_splits is None
    ):
        exit_error(
            "--post-action-num-splits must be defined if --post-action is "
            "mv+sp or cp+sp"
        )
    
    # Sample files if --sample is enabled
    if args.sample:
        if args.sample >= 1.0 and len(files) < int(args.sample):
            exit_error(
                "Not enough files to sample from. The current input has only "
                f"{len(files)} file(s)"
            )
        
        # Sample percentage
        elif args.sample > 0.0 and args.sample < 1.0:
            rng = random.Random(args.random_seed)
            files = rng.sample(files, k=int(args.sample * len(files)))

        # Sample concrete number
        else:
            rng = random.Random(args.random_seed)
            files = rng.sample(files, k=int(args.sample))

    # Collect target files if --post-action
    if args.post_action:
        if (
            args.post_action in ("cp", "mv", "cp+sp", "mv+sp")
            and args.post_action_output is None
        ):
            exit_error(
                "--post-action-output must be defined if --post-action is one "
                "of cp, mv, cp+sp or mv+sp"
            )
        
        post_action_files = []

    # Check --post-action-preserve-subfolders is enabled with --recursive
    if args.post_action_preserve_subfolders and not args.recursive:
        exit_error(
            "--post-action-preserve-subfolders can only be used if --recursive"
            " is enabled"
        )

    # Global stats to collect
    glob_stats = {
        "fs": [],
        "mono_files": 0,
        "stereo_files": 0,
        "multichannel_files": 0,
        "skipped_files": 0,
        "invalid_files": 0,
        "anomalous_files": 0,
        "clipped_files": 0,
        "silent_files": 0,
        "min_duration": None,
        "max_duration": None,
        "total_duration": 0,
        "total_size_bytes": 0
    }

    # Create .csv file if requested
    if args.csv is not None:
        # Header cols (mandatory fields)
        cols = [
            "file",
            "size_bytes",
            "subtype",
            "fmt",
            "fs",
            "num_channels",
            "num_samples_per_channel",
            "duration_seconds",
            "peak_db",
            "rms_db",
            "is_clipped",
            "is_anomalous",
            "is_silent",
            "is_invalid"
        ]

        if args.spectral_rolloff:
            if args.spectral_rolloff_detail:
                for c in (
                    "spectral_rolloff_min",
                    "spectral_rolloff",
                    "spectral_rolloff_max"
                ):
                    cols.insert(-4, c)
            
            else:
                cols.insert(-4, "spectral_rolloff")

        # Optional fields
        if args.sha256 or args.sha256_short:
            cols.insert(1, "sha256")

        with open(args.csv, "w") as f:
            writer = csv.writer(f)
            writer.writerow(cols)
    
    # Mark start
    start_time = perf_counter()

    for file in tqdm(
        files,
        desc="Analyzing audio files",
        colour=get_sppbar_color(),
        leave=False,
        unit="file"
    ):
        # Get metadata
        if args.skip_invalid_files:
            try:
                audio_meta = read_audio_metadata(file)
                audio_meta["file"] = file
                audio_meta["filename"] = os.path.basename(file)
                audio_meta["size_bytes"] = os.path.getsize(file)
                audio_meta["is_invalid"] = False
            
            except Exception as _:
                audio_meta = {}
                audio_meta["file"] = file
                audio_meta["filename"] = os.path.basename(file)
                audio_meta["size_bytes"] = os.path.getsize(file)
                audio_meta["fs"] = None
                audio_meta["num_channels"] = 0
                audio_meta["num_samples_per_channel"] = 0
                audio_meta["duration_seconds"] = 0
                audio_meta["fmt"] = None
                audio_meta["subtype"] = None
                audio_meta["is_invalid"] = True

        else:
            try:
                audio_meta = read_audio_metadata(file)
                audio_meta["file"] = file
                audio_meta["filename"] = os.path.basename(file)
                audio_meta["size_bytes"] = os.path.getsize(file)
                audio_meta["is_invalid"] = False

            except Exception as e:
                exit_error(
                    f"File '{file}' could not be parsed due to the following "
                    f"error: {e}. Use --skip-invalid-files to ignore "
                    "unparseable files and continue analysis"
                )
        
        # Ingest data if requested
        if not args.meta:
            # Skip long files
            if audio_meta["duration_seconds"] > args.max_duration:
                glob_stats["skipped_files"] += 1
                continue

            # Update audio stats
            if args.skip_invalid_files:
                try:
                    audio, fs = read_audio(file, dtype=args.dtype)
                    audio_peak_db = flatten_nested_list(
                        peak_db(audio, axis=-1).tolist()
                    )
                    audio_rms_db = flatten_nested_list(
                        rms_db(audio, axis=-1).tolist()
                    )
                    audio_is_clipped = is_clipped(audio)
                    audio_is_anomalous = is_anomalous(audio)
                    audio_is_silent = is_silent(audio)
                    audio_meta["peak_db"] = audio_peak_db
                    audio_meta["rms_db"] = audio_rms_db
                    audio_meta["is_clipped"] = audio_is_clipped
                    audio_meta["is_anomalous"] = audio_is_anomalous
                    audio_meta["is_silent"] = audio_is_silent
                    audio_meta["is_invalid"] = False

                    if args.sha256 or args.sha256_short:
                        audio_meta["sha256"] = generate_sha256_from_file(file)
                
                except Exception as _:
                    audio_is_clipped = False
                    audio_is_anomalous = False
                    audio_is_silent = False
                    audio_meta["peak_db"] = None
                    audio_meta["rms_db"] = None
                    audio_meta["is_clipped"] = False
                    audio_meta["is_anomalous"] = False
                    audio_meta["is_silent"] = False
                    audio_meta["is_invalid"] = True

                    if args.spectral_rolloff is not None:
                        if args.spectral_rolloff_detail:
                            _spectral_rolloff = spectral_rolloff(
                                audio,
                                fs,
                                args.fft_size,
                                args.hop_size,
                                rolloff=args.spectral_rolloff
                            )
                            _min_spectral_rolloff = flatten_nested_list(
                                np.min(
                                    _spectral_rolloff,
                                    axis=-1,
                                    keepdims=True
                                ).tolist()
                            )
                            _max_spectral_rolloff = flatten_nested_list(
                                    np.max(
                                    _spectral_rolloff,
                                    axis=-1,
                                    keepdims=True
                                ).tolist()
                            )
                            _average_spectral_rolloff = flatten_nested_list(
                                    np.mean(
                                    spectral_rolloff(
                                        audio,
                                        fs,
                                        args.fft_size,
                                        args.hop_size,
                                        rolloff=args.spectral_rolloff
                                    ),
                                    axis=-1,
                                    keepdims=True
                                ).tolist()
                            )
                            audio_meta["spectral_rolloff_min"] = (
                                _min_spectral_rolloff
                            )
                            audio_meta["spectral_rolloff"] = (
                                _average_spectral_rolloff
                            )
                            audio_meta["spectral_rolloff_max"] = (
                                _max_spectral_rolloff
                            )
                        
                        else:
                            audio_meta["spectral_rolloff"] = (
                                    flatten_nested_list(
                                        np.mean(
                                        spectral_rolloff(
                                            audio,
                                            fs,
                                            args.fft_size,
                                            args.hop_size,
                                            rolloff=args.spectral_rolloff
                                        ),
                                        axis=-1,
                                        keepdims=True
                                    ).tolist()
                                )
                            )

                    if args.sha256 or args.sha256_short:
                        audio_meta["sha256"] = generate_sha256_from_file(file)
            
            else:
                try:
                    audio, fs = read_audio(file, dtype=args.dtype)
                    audio_peak_db = flatten_nested_list(
                        peak_db(audio, axis=-1).tolist()
                    )
                    audio_rms_db = flatten_nested_list(
                        rms_db(audio, axis=-1).tolist()
                    )
                    audio_is_clipped = is_clipped(audio)
                    audio_is_anomalous = is_anomalous(audio)
                    audio_is_silent = is_silent(audio)
                    audio_meta["peak_db"] = audio_peak_db
                    audio_meta["rms_db"] = audio_rms_db
                    audio_meta["is_clipped"] = audio_is_clipped
                    audio_meta["is_anomalous"] = audio_is_anomalous
                    audio_meta["is_silent"] = audio_is_silent

                    if args.spectral_rolloff is not None:
                        if args.spectral_rolloff_detail:
                            _spectral_rolloff = spectral_rolloff(
                                audio,
                                fs,
                                args.fft_size,
                                args.hop_size,
                                rolloff=args.spectral_rolloff
                            )
                            _min_spectral_rolloff = flatten_nested_list(
                                np.min(
                                    _spectral_rolloff,
                                    axis=-1,
                                    keepdims=True
                                ).tolist()
                            )
                            _max_spectral_rolloff = flatten_nested_list(
                                    np.max(
                                    _spectral_rolloff,
                                    axis=-1,
                                    keepdims=True
                                ).tolist()
                            )
                            _average_spectral_rolloff = flatten_nested_list(
                                    np.mean(
                                    spectral_rolloff(
                                        audio,
                                        fs,
                                        args.fft_size,
                                        args.hop_size,
                                        rolloff=args.spectral_rolloff
                                    ),
                                    axis=-1,
                                    keepdims=True
                                ).tolist()
                            )
                            audio_meta["spectral_rolloff_min"] = (
                                _min_spectral_rolloff
                            )
                            audio_meta["spectral_rolloff"] = (
                                _average_spectral_rolloff
                            )
                            audio_meta["spectral_rolloff_max"] = (
                                _max_spectral_rolloff
                            )
                        
                        else:
                            audio_meta["spectral_rolloff"] = (
                                    flatten_nested_list(
                                        np.mean(
                                        spectral_rolloff(
                                            audio,
                                            fs,
                                            args.fft_size,
                                            args.hop_size,
                                            rolloff=args.spectral_rolloff
                                        ),
                                        axis=-1,
                                        keepdims=True
                                    ).tolist()
                                )
                            )

                    if args.sha256 or args.sha256_short:
                        audio_meta["sha256"] = generate_sha256_from_file(file)
                
                except Exception as e:
                    exit_error(
                        f"File '{file}' could not be parsed due to the "
                        f"following error: {e}. Use --skip-invalid-files to "
                        "ignore unparseable files and continue analysis"
                    )
            
            # Apply filters
            if (
                (
                    args.filter is not None
                    and _matches_filter(
                        data=audio_meta,
                        preload=preload,
                        expr=args.filter
                    )
                ) or (
                    args.select is not None
                    and not _matches_filter(
                        data=audio_meta,
                        preload=preload,
                        expr=args.select,
                    )
                )
            ):
                continue
  
            if not args.summary:
                file_repr = _audio_file_repr_from_dict(
                    audio_meta,
                    args.max_fname_chars,
                    abbrev_hash=bool(args.sha256_short)
                )
                print(file_repr, writer=tqdm)

            # Collect files for --post-action if any 
            if args.post_action:
                post_action_files.append(file)
        
        else:
            # Format current file representation
            if not args.summary:
                file_repr = _audio_file_meta_repr_from_dict(
                    audio_meta,
                    args.max_fname_chars
                )
                print(file_repr, writer=tqdm)
        
        # Write data to csv
        if args.csv:
            with open(args.csv, "a") as f:
                writer = csv.DictWriter(f, fieldnames=cols)

                # Remove fields not written to .csv
                del audio_meta["filename"]

                writer.writerow(audio_meta)
        
        # Update global stats based on metadata
        if isinstance(audio_meta["duration_seconds"], Number):
            # NOTE: It may not be a number in invalid files
            glob_stats["total_duration"] += audio_meta["duration_seconds"]
        
        # Update size
        glob_stats["total_size_bytes"] += audio_meta["size_bytes"]

        # Update duration stats
        if (
            (glob_stats["min_duration"] is None)
            or (audio_meta["duration_seconds"] < glob_stats["min_duration"])
        ):
            glob_stats["min_duration"] = audio_meta["duration_seconds"]
        
        if (
            (glob_stats["max_duration"] is None)
            or (audio_meta["duration_seconds"] > glob_stats["max_duration"])
        ):
            glob_stats["max_duration"] = audio_meta["duration_seconds"]
        
        # Update channel stats
        if audio_meta["num_channels"] == 1:
            glob_stats["mono_files"] += 1
            
        elif audio_meta["num_channels"] == 2:
            glob_stats["stereo_files"] += 1
        
        elif audio_meta["num_channels"] > 2:
            glob_stats["multichannel_files"] += 1
        
        # Update sample rates
        if audio_meta["fs"] not in glob_stats["fs"]:
            glob_stats["fs"].append(audio_meta["fs"])
        
        # Update global stats based on audio data
        if not args.meta:
            if audio_is_silent:
                glob_stats["silent_files"] += 1
            
            if audio_is_anomalous:
                glob_stats["anomalous_files"] += 1
            
            if audio_is_clipped:
                glob_stats["clipped_files"] += 1
            
            if audio_meta["is_invalid"]:
                glob_stats["invalid_files"] += 1
            
    # Get elapsed time
    elapsed_time = perf_counter() - start_time

    # Print global stats
    if not args.summary:
        print("")
    
    print(
        "Total file(s):".ljust(22) + str(
            glob_stats["mono_files"]
            + glob_stats["stereo_files"]
            + glob_stats["multichannel_files"]
            + glob_stats["invalid_files"]
        )
    )

    if glob_stats["invalid_files"] > 0:
        print_error(
            "Invalid file(s):".ljust(22) + f"{glob_stats['invalid_files']}"
        )

    print("Mono file(s):".ljust(22) + f"{glob_stats['mono_files']}")
    print("Stereo file(s):".ljust(22) + f"{glob_stats['stereo_files']}")
    print(
        "Multichannel file(s):".ljust(22)
        + f"{glob_stats['multichannel_files']}",
    )

    if len(glob_stats["fs"]) == 0:
        fs_repr = "-"

    else:
        fs_repr = ", ".join(
            f"{fs}hz" if fs is not None
            else "unknown" for fs in glob_stats["fs"]
        )

    if "unknown" in fs_repr:
        print_error("Sample rate(s):".ljust(22) + f"{fs_repr}")
    
    else:
        print("Sample rate(s):".ljust(22) +  f"{fs_repr}")
    
    # Data dependant summary lines
    if not args.meta:
        skipped_files_repr = (
            "Skipped files:".ljust(22) + str(glob_stats["skipped_files"])
        )

        if glob_stats["skipped_files"] > 0:
            print_error(skipped_files_repr)
        
        else:
            print(skipped_files_repr)

        clipped_files_repr = (
            "Clipped files:".ljust(22) + str(glob_stats["clipped_files"])
        )

        if glob_stats["clipped_files"] > 0:
            print_error(clipped_files_repr)

        else:
            print(clipped_files_repr)

        anomalous_files_repr = (
            "Anomalous files:".ljust(22) + str(glob_stats["anomalous_files"])
        )

        if glob_stats["anomalous_files"] > 0:
            print_error(anomalous_files_repr)
        
        else:
            print(anomalous_files_repr)

        silent_files_repr = (
            "Silent files:".ljust(22) + str(glob_stats["silent_files"])
        )

        if glob_stats["silent_files"] > 0:
            print_error(silent_files_repr)
        
        else:
            print(silent_files_repr)

    print(
        "Total duration:".ljust(22)
        + time_to_str(glob_stats['total_duration']),
    )

    # NOTE: Total files are recalculated because some files may have been
    # filtered from len(files)
    total_files = (
        glob_stats["mono_files"]
        + glob_stats["stereo_files"]
        + glob_stats["multichannel_files"]
    )

    if total_files > 1:
        print(
            "Minimum duration:".ljust(22)
            + time_to_str(glob_stats['min_duration'])
        )
        print(
            "Maximum duration:".ljust(22)
            + time_to_str(glob_stats['max_duration'])
        )
        print(
            "Average duration:".ljust(22)
            + time_to_str(glob_stats['total_duration'] / total_files)
        )

    print(
        "Total size:".ljust(22) + bytes_to_str(glob_stats['total_size_bytes']),
    )
    print("")
    print(f"Elapsed time: {time_to_str(elapsed_time, abbrev=True)}")

    # Perform --post-action if any
    if args.post_action:
        _perform_post_action(post_action_files, args)



import numpy as np
from typing import Optional
from scipy.signal import stft
from .config import get_default_eps
from .collections import flatten_nested_list


def amp_to_db(x: np.ndarray, eps: float = get_default_eps()) -> np.ndarray:
    """Returns the amplitude in decibels of every sample.
    
    Args:
        x (np.ndarray): Input audio data.
        eps (float): Machine epsilon.
    
    Returns:
        np.ndarray: Array containing values in decibel scale.
    """
    return 20.0 * np.log10(np.clip(x, a_min=eps, a_max=None))


def db_to_amp(x: np.ndarray, min: Optional[float] = None) -> np.ndarray:
    """Transforms decibel values to amplitude values.
    
    Args:
        x (np.ndarray): Array containing decibel values.
        min (Optional[float]): Minimum decibel value. Values below this
            threshold will be replaced by 0.0 to eliminate denormals.
    
    Returns:
        np.ndarray: Array containing amplitude values.
    """
    amp = 10.0 ** (x / 20.0)

    if min is not None:
        amp = np.where(amp <= min, 0.0, amp)
    
    return amp


def peak_amp(x: np.ndarray, axis: int = -1) -> np.ndarray:
    """Returns the peak amplitude of a `np.ndarray`.
    
    Args:
        x (np.ndarray): Input audio data.
        axis (int): Axis along which the peak amplitude is calculated.
    
    Returns:
        np.ndarray: Array containing peak amplitude values.
    """
    return np.amax(np.abs(x), axis=axis, keepdims=True)


def peak_db(
        x: np.ndarray,
        axis: int = -1,
        eps: float = get_default_eps()
) -> np.ndarray:
    """ Returns the peak amplitude of a `np.ndarray` in decibel scale.
    
    Args:
        x (np.ndarray): Input audio data.
        axis (int): Axis along which the peak amplitude is calculated.
        eps (float): Machine epsilon.
    
    Returns:
        np.ndarray: Array containing peak amplitude values in decibel scale.
    """
    return amp_to_db(peak_amp(x, axis=axis), eps=eps)


def rms(x: np.ndarray, axis: int = -1) -> np.ndarray:
    """Returns the root mean square level of a `np.ndarray`.
    
    Args:
        x (np.ndarray): Input audio data.
        axis (int): Axis along which the root mean square level is calculated.
    
    Returns:
        np.ndarray: Array containing the root mean square level.
    """
    return np.mean(x ** 2, axis=axis, keepdims=True) ** 0.5


def rms_db(x: np.ndarray, axis: int = -1) -> np.ndarray: 
    """ Returns the root mean square level of a `np.ndarray` in decibel scale.
    
    Args:
        x (np.ndarray): Input audio data.
        axis (int): Axis along which the root mean square level is calculated.
    
    Returns:
        np.ndarray: Array containing the root mean square level in decibels.
    """
    return amp_to_db(rms(x, axis=axis))


def is_clipped(x: np.ndarray, min: float = -1.0, max: float = 1.0) -> bool:
    """Returns `True` if a `np.ndarray` containing audio data has values
    above or below a certain range.

    Args:
        x (np.ndarray): Input audio data.
        min (float): Lower limit of the allowed range.
        max (float): Upper limit of the allowed range.
    
    Returns:
        bool: `True` if the file contains at least one value outside the given
            range.
    """
    result = np.logical_or(np.any(x > max), np.any(x < min))
    return bool(result)


def is_anomalous(x: np.ndarray) -> bool:
    """ Retruns `True`if a `np.ndarray` containing audio data has `inf`, `-inf`
    or `NaN` values.
    
    Args:
        x (np.ndarray): Input audio data.
    
    Returns:
        bool: `True` if the file contains at least one `inf`, `-inf` or `NaN`
            value.
    """
    result = np.logical_or(
        np.logical_or(np.isinf(x).any(), np.isneginf(x).any()),
        np.isnan(x).any()
    )
    return bool(result)

def is_silent(x: np.ndarray, thresh_db: float = -80.0, axis: int = -1) -> bool:
    """Returns `True` if a `np.ndarray` containing audio data is silent. That
    is, the root mean square level of the files in decibels is below a certain
    threshold.

    Args:
        x (np.ndarray): Input audio data.
        thresh_db (float): Minimum threshold below which a file is considered
            silent.
        axis (int): Axis along which the root mean square level in decibels is
            computed and contrasted again the given threshold in decibels.
    
    Returns:
        bool: `True` if `x` is silent, `False` otherwise.
    """
    db_rms = flatten_nested_list(rms_db(x, axis=axis).tolist())
    return all(chl_db_rms < thresh_db for chl_db_rms in db_rms)


def spectral_rolloff(
        x: np.ndarray,
        fs: int,
        fft_size: int,
        hop_size: Optional[int],
        window: str = "hann",
        rolloff: float = 0.9
) -> np.ndarray:
    """Calculates the spectral rolloff of an input array `x`. That is, the
    frequency bin under which `rolloff` percent of the energy is
    accumulated.

    Args:
        x (np.ndarray): Input audio data.
        fs (int): Sample rate.
        fft_size (int): Size of the FFT.
        hop_size (Optional[int]): Hop size of the FFT.
        window (str): Window type.
        rolloff (float): Rolloff percent between 0.0 and 1.0. Rolloff of
            0.9 means that the resulting rolloff for a given frequency is
            the value under which 90 percent of the energy is accumulated.
    
    Returns:
        np.ndarray: Array containing framewise roll-off.
    """
    if rolloff < 0.0 or rolloff > 1.0:
        raise ValueError("rolloff must be between 0.0 and 1.0")
    
    # Compute magnitude
    fc, _, x_stft = stft(
        x,
        fs=fs,
        nperseg=fft_size,
        nfft=None,
        noverlap=fft_size - hop_size,
        window=window,
        return_onesided=True,
        padded=False,
        scaling="spectrum"
    )
    x_mag = np.abs(x_stft)
    fc = np.expand_dims(fc, axis=(0, -1))
    fc = np.broadcast_to(fc, x_mag.shape)

    # Get cumulative sum and obtain rolloff threshold per frame
    x_mag_cumsum = np.cumsum(x_mag, axis=-2)
    rolloff_threshold = np.expand_dims(
        rolloff * x_mag_cumsum[..., -1, :], axis=-2
    )

    # Mask all values below threshold as inf
    rolloff_idx = np.where(x_mag_cumsum < rolloff_threshold, np.nan, 1.0)
    rolloff_freq = np.nanmin(fc * rolloff_idx, axis=-2, keepdims=True)

    return rolloff_freq



import os
import numpy as np
import soundfile as sf
from glob import glob
from typing import (
    Callable,
    List,
    Optional,
    Tuple,
    Union
)
from .config import get_default_audio_io_dtype
from .exceptions import FolderNotFoundError
from .guards import is_file_or_error
from .collections import make_list
from .fmt import (
    _decorate_str,
    exit_warning,
    print_error
)


def get_dir_files(
        dir: Union[str, List[str]],
        ext: Union[str, List[str]] = ".wav",
        recursive: bool = True,
        key: Optional[Callable] = None,
) -> List[str]:
    """Returns a `list` with all the files inside folder with extension `ext`.
    It supports a recursive search and searching in more than one root folder
    at a time if `recursive=True` and `dir` is a `list` of `str`,
    respectively.

    Args:
        dir (Union[str, List[str]]): Folder(s) to be searched.
        ext (Union[str, Tuple[str]]): File extensions to be considered. Accepts
            `.*` as a wild card.
        recursive (bool): If `True`, the search inside each folder will be
            recursive.
        key (Optional[Callable]): Key function to sort the results. If it is
            not provided, files will be sorted alphabetically.

    Returns:
        `list` of `str` with the path to each retrieved file.

    Raises:
        FileNotFoundError: If one of the folder(s) cannot be found.
    """
    dir = make_list(dir)
    ext = make_list(ext)

    # Check dirs exist before fetching content
    for dir_ in dir:
        if not os.path.isdir(dir_):
            raise FolderNotFoundError(f"Folder not found: '{dir_}'")

    all_files = []

    # Search dirs
    for dir_ in dir:
        for ext_ in ext:
            if recursive:
                all_files.extend(
                    list(
                        glob(os.path.join(dir_, "**", f"*{ext_}"),
                             recursive=True)
                    )
                )
            else:
                all_files.extend(list(glob(os.path.join(dir_, f"*{ext_}"))))
    
    # Filter out f olders with file-like names (e.g. ending in .wav extension)
    flagged_files = []

    for file in all_files:
        if not os.path.isfile(file):
            flagged_files.append(file)
    
    for flagged_file in flagged_files:
        all_files.remove(flagged_file)

    return sorted(all_files, key=key)


def read_audio_metadata(file: str) -> dict:
    """Reads the metadata block from an audio files and returns it as a
    `dict`.

    Args:
        file (str): Audio file.
    
    Returns:
        (dict): Audio metadata.
    """
    meta = sf.info(file, verbose=False)

    return {
        "fs": meta.samplerate,
        "num_channels": meta.channels,
        "num_samples_per_channel": meta.frames,
        "duration_seconds": meta.duration,
        "fmt": meta.format,
        "subtype": meta.subtype
    }


def read_audio(
        file: str,
        start: int = 0,
        frames: Optional[int] = -1,
        stop: Optional[int] = None,
        dtype: str = get_default_audio_io_dtype(),
) -> Tuple[np.ndarray, int]:
    """Reads an audio file or audio file chunk and returns it as a 
    `np.ndarray`.

    Args:
        file (str): Audio file.
        start (int): Start frame for reading partial frames of the file.
        frames Optional[int]: Number of frames to read.
        stop (Optional[int]): End frame index for reading partial frames of the
            file.
        dtype (str): Data type used to represent the data.
    
    Returns:
        (Tuple[np.ndarray, int]): `np.ndarray` representing the audio data and
            and sample rate `tuple`.
    """
    is_file_or_error(file)

    # Read audio in (num_channels, num_samples) format
    data, fs_ = sf.read(
        file,
        dtype=dtype,
        always_2d=True,
        start=start,
        stop=stop,
        frames=frames
    )

    return data.transpose(), fs_


def ask_confirmation(
            s: str = "<magenta><b>Do you want to continue? [y/n]:</b>"
                     "</magenta> ",
            exit_s: str = "Program finished by the user",
            exit: bool = True
    ) -> Optional[bool]:
        """Request user input to confirm or reject an instruction.

        Args:
            s (str): Message to be printed to ask user confirmation.
            exit (bool): If `True` and user answer is `n` (no), then
                the program execution is terminated.

        Returns:
            Optional[bool]: User response.
        """
        user_input = None

        while str(user_input) not in ["y", "n"]:
            if user_input is not None:
                print_error(f"Invalid input '{user_input}'")

            user_input = input(_decorate_str(s))

            if str(user_input) == "y":
                response = True

            elif str(user_input) == "n":
                if exit:
                    exit_warning(exit_s)

                else:
                    response = False
            else:
                pass

        return response



import os
from typing import (
    List,
    Union
)
from .exceptions import FileExtensionError
from .collections import make_list


def is_file_or_error(file: str) -> None:
    """Raises and exception if `file` is not a file.

    Args:
        file (str): File path to check.

    Raises:
        FileNotFoundError: If `file` is not a valid file.
    """
    if not os.path.isfile(file):
        raise FileNotFoundError(f"File not found: {file}")


def has_ext(file: str, ext: Union[str, List[str]]) -> bool:
    """Returns `True` if a file has a certain extension.
    
    Args:
        file (str): File to check.
        ext (Union[str, List[str]]): Single extension to check as `str` or
            `list` of extensions to check.
    
    Returns:
        bool: `True` if `file` has one of the specified extensions, `False`
            otherwise.
    """
    ext = make_list(ext)
    _, ext_ = os.path.splitext(file)
    return ext_ in ext


def has_ext_or_error(file: str, ext: Union[str, List[str]]) -> None:
    """Raises an exception if a file does not have an extension among a set
    of specified extensions.
    
    Args:
        file (str): File to check.
        ext (Union[str, List[str]]): Single extension to check as `str` or
            `list` of extensions to check.
    
    Raises:
        FileExtensionError: If `file` does not have any of the specified
            extensions.
    """
    if not has_ext(file, ext=ext):
        ext_repr = ", ".join([f"'{e}'" for e in ext])

        raise FileExtensionError(
            f"Invalid file extension of '{file}'. Expected file extensions: "
            f"{ext_repr}"
        )


def is_file_with_ext(file: str, ext: Union[str, List[str]]) -> bool:
    """Returns `True` if `file` exists and has one of the specified
    extensions.
    
    Args:
        file (str): File to check.
        ext (Union[str, List[str]]): Single extension to check as `str` or
             `list` of extensions to check.
    
    Returns:
        bool: `True` if `file` exists and has one of the specified extensions,
            `False` otherwise.
    """
    return os.path.isfile(file) and has_ext(file, ext)


def is_file_with_ext_or_error(file: str, ext: Union[str, List[str]]) -> bool:
    """Raises an exception if `file` does not exists or does not have one of
    the specified extensions.
    
    Args:
        file (str): File to check.
        ext (Union[str, List[str]]): Single extension to check as `str` or
            `list` of extensions to check.
    """
    is_file_or_error(file)
    has_ext_or_error(file, ext=ext)



import sys
from tqdm import tqdm
from typing import Optional
from .config import (
    _get_text_color_tags,
    _get_text_decorator_tags
)


def _decorate_str(s: str) -> str:
    """Replaces colors and decorators in a string.

    Args:
        s (str): The input string to be decorated.

    Returns:
        str: The decorated string.
    """
    # Replace colors and decorators
    for k, v in _get_text_decorator_tags().items():
        s = s.replace(k, v)

    for k, v in _get_text_color_tags().items():
        s = s.replace(k, v)

    return s


def printc(s: str, writer: Optional[tqdm] = None) -> None:
    """Prints a formatted string.

    Args:
        s (str): The string to print.
        writer (Optional[tqdm]): Writer to use.
    """
    return (
        print(_decorate_str(s)) if writer is None
        else writer.write(_decorate_str(s))
    )


def printc_exit(
        s: str,
        code: int = 0,
        writer: Optional[tqdm] = None
) -> None:
    """Prints a formatted string and exits the program with a specified exit
    code.

    Args:
        s (str): The string to print.
        code (int): Exit code.
        writer (Optional[tqdm]): Writer to use.
    """
    printc(s=s, writer=writer)
    sys.exit(code)


def print_error(s: str, writer: Optional[tqdm] = None) -> None:
    """Prints an error.
    
    Args:
        s (str): Error message print.
        writer (Optional[tqdm]): Writer to use.
    """
    printc(f"<error>{s}</error>", writer=writer)


def print_warning(s: str, writer: Optional[tqdm] = None) -> None:
    """Prints a warning.
    
    Args:
        s (str): Warning message to print.
        writer (Optional[tqdm]): Writer to use.
    """
    printc(f"<warning>{s}</warning>", writer=writer)


def exit_error(s: str, code: int = 1, writer: Optional[tqdm] = None) -> None:
    """Prints an error and stops the execution of the program.
    
    Args:
        s (str): Error message to print.
        code (int): Exit code.
        writer (Optional[tqdm]): Writer to use.
    """
    print_error(s, writer=writer)
    sys.exit(code)


def exit_warning(s: str, code: int = 1, writer: Optional[tqdm] = None) -> None:
    """Prints a warning and stops the execution of the program.

    Args:
        s (str): Warning message to print.
        code (int): Exit code.
        writer (Optional[tqdm]): Writer to use.
    """
    print_warning(s, writer=writer)
    sys.exit(code)


def bytes_to_str(bytes: int) -> str:
    """Returns an amount of bytes in a human readable format.
    
    Args:
        bytes (int): Number of bytes to represent.

    Returns:
        (str): `str` representation of `bytes`.
    """
    if bytes / (1024 ** 4) > 1.0:
        repr = f"{bytes / 1024 ** 4:.1f}T"

    elif bytes / (1024 ** 3) > 1.0:
        repr = f"{bytes / 1024 ** 3:.1f}G"
        
    elif bytes / (1024 ** 2) > 1.0:
        repr = f"{bytes / 1024 ** 2:.1f}M"
        
    elif bytes / 1024 > 1.0:
        repr = f"{bytes / 1024:.1f}K"
        
    else:
        repr = f"{bytes:d}B"
    
    return repr


def time_to_str(time: float, abbrev: bool = False) -> str:
    """Returns a time in seconds in a human readable format.
    
    Args:
        time (float): Time in seconds.
        abbrev (bool): If `True`, abbreviations to represent different time
            units are used.
    
    Returns:
        (str): `str` representation of `time`.
    """
    if abbrev:
        ms_repr = "ms"
        s_repr = "s"
        m_repr = "m"
        h_repr = "h"
    
    else:
        ms_repr = "millisecond(s)"
        s_repr = "second(s)"
        m_repr = "minute(s)"
        h_repr = "hour(s)"

    if time < 1.0:
        time_repr = f"{time * 1e3:.1f} {ms_repr}"

    elif time < 60.0:
        time_repr = f"{time:.1f} {s_repr}"
    
    elif time < 3600.0:
        time_mins = time // 60.0
        time_secs = time % 60
        time_repr = f"{int(time_mins)} {m_repr} {int(time_secs)} {s_repr}" 

    else:            
        time_hours = time // 3600.0
        remaining_time = time % 3600.0
        time_mins = remaining_time // 60.0
        time_secs = remaining_time % 60
        time_repr = (
            f"{int(time_hours)} {h_repr} {int(time_mins)} {m_repr} "
            f"{int(time_secs)} {s_repr}"
        )

    return time_repr



"""An audio-friendly ls, with a little something extra"""
from datetime import datetime

__description__ = "an audio-friendly ls, with a little something extra"
__version__ = "0.1.7"
__version_repr__ = (
    f"sndls version {__version__} 2025-{datetime.now().year} developed by "
    "Esteban Gómez (Speech Interaction Technology, Aalto University)"
)



from itertools import chain
from typing import (
    Any,
    List
)


def make_list(x: Any) -> List[Any]:
    """If `x` is a single element, turns it into a `list` of one element.

    Args:
        x (Any): Element(s) to be returned as a `list`.

    Returns:
        (list): `x` as a `list`.
    """
    return [x] if not isinstance(x, list) and x is not None else x


def flatten_nested_list(nl: List[List]) -> List[Any]:
    """Flattens a list of lists of arbitrary depth.

    Args:
        nl (List[List]): Nested `list` of arbitrary depth.

    Return:
        (list): Flattened `list`.
    """
    return (
        list(chain.from_iterable(map(flatten_nested_list, nl)))
        if isinstance(nl, list)
        else [nl]
    )


def time_to_str(time: float, abbrev: bool = False) -> str:
    """ Returns a time in seconds in a human readable format.
    
    Args:
        time (float): Time in seconds.
        abbrev (bool): If `True`, abbreviations to represent different time
            units are used.
    
    Returns:
        (str): `str` representation of `time`.
    """
    if abbrev:
        ms_repr = "ms"
        s_repr = "s"
        m_repr = "m"
        h_repr = "h"
    
    else:
        ms_repr = "millisecond(s)"
        s_repr = "second(s)"
        m_repr = "minute(s)"
        h_repr = "hour(s)"

    if time < 1.0:
        time_repr = f"{time * 1e3:.1f} {ms_repr}"

    elif time < 60.0:
        time_repr = f"{time:.1f} {s_repr}"
    
    elif time < 3600.0:
        time_mins = time // 60.0
        time_secs = time % 60
        time_repr = f"{int(time_mins)} {m_repr} {int(time_secs)} {s_repr}" 

    else:            
        time_hours = time // 3600.0
        remaining_time = time % 3600.0
        time_mins = remaining_time // 60.0
        time_secs = remaining_time % 60
        time_repr = (
            f"{int(time_hours)} {h_repr} {int(time_mins)} {m_repr} "
            f"{int(time_secs)} {s_repr}"
        )

    return time_repr


def bytes_to_str(bytes: int) -> str:
    """Returns an amount of bytes in a human readable format.
    
    Args:
        bytes (int): Number of bytes to represent.

    Returns:
        (str): `str` representation of `bytes`.
    """
    if bytes / (1024 ** 4) > 1.0:
        repr = f"{bytes / 1024 ** 4:.1f}T"

    elif bytes / (1024 ** 3) > 1.0:
        repr = f"{bytes / 1024 ** 3:.1f}G"
        
    elif bytes / (1024 ** 2) > 1.0:
        repr = f"{bytes / 1024 ** 2:.1f}M"
        
    elif bytes / 1024 > 1.0:
        repr = f"{bytes / 1024:.1f}K"
        
    else:
        repr = f"{bytes:d}B"
    
    return repr



import numpy as np
from typing import Tuple


class __Singleton__(type):
    """A singleton class to be used as `metaclass`."""
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class __Config__(metaclass=__Singleton__):
    """Internal package configuration singleton.

    !!! warning
        This singleton is not supposed to be accessed directly. Please use the
        respective setters and getters for each configuration.
    """
    def __init__(self):
        super().__init__()

        self._TEXT_COLORS = {
            "red": "\033[91m",
            "green": "\033[92m",
            "blue": "\033[94m",
            "cyan": "\033[96m",
            "magenta": "\033[35m",
            "yellow": "\033[93m",
            "end_color": "\033[0m",
        }

        self._TEXT_DECORATORS = {
            "bold": "\033[1m",
            "italic": "\033[3m",
            "underline": "\033[4m",
            "end_decoration": "\033[0m",
        }

        self._TEXT_COLOR_TAGS = {
            "<error>": self._TEXT_COLORS["red"],
            "</error>": self._TEXT_COLORS["end_color"],
            "<warning>": self._TEXT_COLORS["yellow"],
            "</warning>": self._TEXT_COLORS["end_color"],
            "<green>": self._TEXT_COLORS["green"],
            "</green>": self._TEXT_COLORS["end_color"],
            "<magenta>": self._TEXT_COLORS["magenta"],
            "</magenta>": self._TEXT_COLORS["end_color"]
        }

        self._TEXT_DECORATOR_TAGS = {
            "<b>": self._TEXT_DECORATORS["bold"],
            "</b>": self._TEXT_DECORATORS["end_decoration"],
            "<i>": self._TEXT_DECORATORS["italic"],
            "</i>": self._TEXT_DECORATORS["end_decoration"],
            "<u>": self._TEXT_DECORATORS["underline"],
            "</u>": self._TEXT_DECORATORS["end_decoration"]
        }

        self._DEFAULT_AUDIO_IO_DTYPE = "float32"
        self._DEFAULT_AUDIO_SUBTYPE = "FLOAT"
        self._SINGLE_PROCESS_PROGRESS_BAR_COLOR = "green"
        self._MULTI_PROCESS_PROGRESS_BAR_COLOR = "cyan"
        self._ALLOWED_AUDIO_FILE_EXTENSIONS = (
            ".aif",
            ".aiff",
            ".mp3",
            ".flac",
            ".ogg",
            ".wav",
            ".wave"
        )


def _get_text_color_tags() -> dict:
    """Returns all available text color tags.
    
    Returns:
        dict: Text color tags.
    """
    return __Config__()._TEXT_COLOR_TAGS


def _get_text_decorator_tags() -> dict:
    """Returns all available decorator tags.
    
    Returns:
        dict: Decorator tags.
    """
    return __Config__()._TEXT_DECORATOR_TAGS


def get_sppbar_color() -> str:
    """Returns the default single process progress bar color.
    
    Returns:
        str: Single process progress bar color.
    """
    return __Config__()._SINGLE_PROCESS_PROGRESS_BAR_COLOR


def get_datetime_fmt() -> str:
    """Returns the default date and time format.
    
    Returns:
        str: Default date and time format.
    """
    return __Config__()._DATETIME_FMT


def get_default_audio_io_dtype() -> str:
    """Returns the default data type used to read audio files.
    
    Returns:
        str: Default data type used to read audio files.
    """
    return __Config__()._DEFAULT_AUDIO_IO_DTYPE


def get_default_eps() -> float:
    """Returns the default epsilon value for the current default audio read
    data type.
    
    Returns:
        float: Default epsilon for the current default data type used to read
            audio data.
    """
    return np.finfo(get_default_audio_io_dtype()).eps


def get_allowed_audio_file_extensions() -> Tuple[str]:
    """Returns the allowed audio file extensions.
    
    Returns:
        Tuple(str): Allowed audio file extensions.
    """
    return __Config__()._ALLOWED_AUDIO_FILE_EXTENSIONS
